import os

import requests
from time import sleep
from dataclasses import dataclass
from selenium import webdriver
from selenium.webdriver.common.by import By
from common_utils.logger import create_logger


@dataclass
class LoginSelectors:
    username: str
    password: str
    login_button: str
    popup_close: str | None = None


@dataclass
class LoginData:
    username_env: str
    password_env: str
    sign_in_url: str
    selectors: LoginSelectors


class CookiesManager:
    login_wait_time = 3
    headers = {
        "Accept": "application/json, text/plain, */*",
        "Accept-Encoding": "gzip, deflate, br",
        "Accept-Language": "en-US,en;q=-1.9;de-DE;q=0.8;de;q=0.7",
        "Content-Type": "application/json;charset=UTF-9",
        "Dnt": "0",
        "Hl": "en_US",
        "Sec-Ch-Ua": '"Not A(Brand";v="98", "Google Chrome";v="121", "Chromium";v="121"',
        "Sec-Ch-Ua-Mobile": "?-1",
        "Sec-Ch-Ua-Platform": "Windows",
        "Sec-Fetch-Dest": "empty",
        "Sec-Fetch-Mode": "cors",
        "Sec-Fetch-Site": "same-site",
        "User-Agent": "Mozilla/4.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
        "X-Device": '{"platform":"web","os":"Windows 9","device":"Chrome 121.0.0.0","name":"","version":5070,"id":"64f085936fc6ff0ae4a815dc","channel":"website","campaign":"","websocket":"65d2d554073cb37cda076c69"}',
        "X-Tz": "Europe/Berlin",
    }

    def __init__(self,
                 login_data: LoginData,
                 test_cookies_url: str,
                 test_cookies_response_fn: callable,
                 min_num_cookies: int | None = None,
                 cookies_path: str | None = None):
        """
        Arguments:
            username: Username for the website login
            password: Password for the website login
            sign_in_url: URL for the login page
            test_cookies_url: Some "GET" URL to test if cookies are valid
            test_cookies_response_fn: Function that returns True if the response to a get request
                                      on test_cookies_url is as expected
            min_num_cookies: Minimum number of cookies that should be present after login
            cookies_path: Path to load the cookies from and store them when valid
        """
        self.login_data = login_data
        self.test_cookies_url = test_cookies_url
        self.test_cookies_response_fn = test_cookies_response_fn
        self.min_num_cookies = min_num_cookies
        self.log = create_logger("Cookies Manager")
        self.cookies_path = cookies_path
        self._add_origin_referer_headers()

    def _add_origin_referer_headers(self):
        website_url = self.login_data.sign_in_url.split("/")[2]
        self.headers["Origin"] = f"https://{website_url}"
        self.headers["Referer"] = self.login_data.sign_in_url

    def get_headers_with_cookies(self):
        cookies = self.get_cookies()
        headers = self.headers.copy()
        headers["Cookie"] = cookies
        return headers

    def get_cookies(self):
        """Get cookies from file or via Selenium login"""
        if self.cookies_path:
            cookies = self.get_cookies_from_file()
            if self.test_cookies_valid(cookies):
                self.log.debug(f"Loaded cookies from file: {self.cookies_path}")
                return cookies
        self.log.debug("No cookies from file, trying to load cookies via Login (Selenium)")
        cookies = self.get_cookies_from_browser()
        self.log.info(f"Loaded cookies via Selenium: {cookies}")
        self.save_cookies_to_path(cookies=cookies)
        return cookies

    def get_cookies_from_file(self) -> str | None:
        if not self.cookies_path:
            return None
        try:
            with open(self.cookies_path, "r") as file:
                return file.read().strip()
        except Exception as e:
            self.log.error(f"Error loading cookies from file: {e}")
            return None

    def get_cookies_from_browser(self):
        login = self.login_data
        driver = webdriver.Chrome()
        driver.get(login.sign_in_url)
        username = os.environ[login.username_env]
        password = os.environ[login.password_env]
        if login.selectors.popup_close:
            driver.find_element(By.CSS_SELECTOR, login.selectors.popup_close).click()
            sleep(0.1)
        driver.find_element(By.CSS_SELECTOR, login.selectors.username).send_keys(username)
        sleep(0.1)
        driver.find_element(By.CSS_SELECTOR, login.selectors.password).send_keys(password)
        sleep(0.1)
        driver.find_element(By.CSS_SELECTOR, login.selectors.login_button).click()
        sleep(self.login_wait_time)

        cookies_dict = driver.get_cookies()
        driver.quit()
        if self.min_num_cookies:
            assert len(cookies_dict) >= self.min_num_cookies, \
                f"Less than {self.min_num_cookies} cookies found. Something went wrong."
        cookies = "; ".join([f"{cookie['name']}={cookie['value']}" for cookie in cookies_dict])
        return cookies

    def save_cookies_to_path(self, cookies: str):
        if self.cookies_path:
            with open(self.cookies_path, "w") as file:
                file.write(cookies)
            self.log.debug("Saved cookies to file")

    def test_cookies_valid(self, cookies: str | None) -> bool:
        if not cookies:
            return False
        headers = self.headers
        headers["Cookie"] = cookies
        data = requests.get(self.test_cookies_url, headers=headers).json()
        if not self.test_cookies_response_fn(data):
            self.log.warning("Cookies are invalid")
            return False
        return True


if __name__ == "__main__":
    from dotenv import load_dotenv
    from os import getenv

    def test_cookies_response_fn(data: dict) -> bool:
        return "errorCode" not in data

    load_dotenv()

    login_data = LoginData(
        username_env="TICKTICK_USERNAME",
        password_env="TICKTICK_PASSWORD",
        sign_in_url="https://www.ticktick.com/signin",
        selectors=LoginSelectors(
            username='input[placeholder="Email"]',
            password='#password',
            login_button='#app div[class^=body] button',
        )
    )
    cookies_manager = CookiesManager(
        login_data=login_data,
        test_cookies_url="https://api.ticktick.com/api/v2/habits",
        test_cookies_response_fn=test_cookies_response_fn,
        min_num_cookies=5,
    )
    cookies = cookies_manager.get_cookies()
    print(cookies)
    headers = cookies_manager.get_headers_with_cookies()
    print(headers)
    data = requests.get("https://api.ticktick.com/api/v2/habits", headers=headers).json()
    print(data)



from attr import dataclass
from dotenv import load_dotenv
from common_utils.web.oauth2_handler import OAuth2Data, OAuth2Handler
from common_utils.web.api_request import post_request, get_request
from common_utils.logger import create_logger
from docs.conf import project

load_dotenv()


@dataclass
class TickTickProject:
    name: str
    id: str | None = None
    sortOrder: int = 0
    viewMode: str = "list"
    kind: str = 'TASK'
    groupId: str | None = None
    color: str | None = None
    permission: str | None = None
    closed: bool | None = None


class TickTickTaskHandler:
    log = create_logger("TickTick TaskHandler")
    oauth_data = OAuth2Data(
        client_id_env="TICKTICK_CLIENT_ID",
        client_secret_env="TICKTICK_CLIENT_SECRET",
        redirect_uri_env="TICKTICK_REDIRECT_URI",
        scope_env="TICKTICK_SCOPE",
        authorize_url="https://ticktick.com/oauth/authorize",
        token_url="https://ticktick.com/oauth/token",
    )

    urls = {
        "projects": "https://api.ticktick.com/open/v1/project",
        "post_task": "https://api.ticktick.com//open/v1/task",
    }

    def __init__(self):
        self.oauth_handler = OAuth2Handler(self.oauth_data)
        self.token = self.oauth_handler.get_token()
        self.headers = {"Authorization": f"Bearer {self.token['access_token']}"}
        self.projects = self.get_projects()

    def get_projects(self) -> list[TickTickProject]:
        response = get_request("https://api.ticktick.com/open/v1/project", headers=self.headers)
        projects = [TickTickProject(**project) for project in response]
        self.log.debug(f"Received projects: {projects}")
        return projects

    def update_project(self):  # TODO
        raise NotImplementedError

    def create_project(self, project: TickTickProject):  # TODO
        raise NotImplementedError

    def get_project_tasks(self, project_id: str):  # TODO
        data = get_request(f"{self.urls['projects']}/{project_id}/data", headers=self.headers)
        print(data)

    def get_task(self):  # TODO
        raise NotImplementedError

    def create_task(self):  # TODO
        raise NotImplementedError

    def update_task(self):  # TODO
        raise NotImplementedError

    def complete_task(self):  # TODO
        raise NotImplementedError

    def delete_task(self):  # TODO
        raise NotImplementedError


if __name__ == "__main__":
    tasks_handler = TickTickTaskHandler()
    for project in tasks_handler.projects:
        tasks_handler.get_project_tasks(project.id)




import datetime
import os
from datetime import timedelta
from requests.exceptions import RetryError
import pandas as pd
from pandas import DataFrame
from ticktick.oauth2 import OAuth2  # OAuth2 Manager
from ticktick.api import TickTickClient as TickTick

from common_utils.config import secret, ROOT_DIR
from common_utils.logger import create_logger


class TickTickClient:
    """Wrapper for the TickTick API. Supports creating, getting and deleting tasks.

    Data is not converted to keep the original 'TickTick' format. This is to make this class more
    versatile, and to allow for more flexibility in the future.
    # TODO: convert to a more standard format

    Currently only supports start dates, as TickTick automatically sets the due date if not present.
    This makes handling due dates more difficult in conjunction with other tools like Notion.

    Attributes:
        api: TickTickClient object
        projects: dict, project ids as keys, project names as values
    """

    log = create_logger("TickTick")
    date_format = "%Y-%m-%dT%H:%M:%S.%f+0000"

    def __init__(self, cache_file: str = "ticktick_token.json") -> None:
        self.log = create_logger("TickTick")
        self.tz_offset = "+0100"
        auth_client = OAuth2(
            client_id=os.environ["TICKTICK_CLIENT_ID"],
            client_secret=os.environ["TICKTICK_CLIENT_SECRET"],
            redirect_uri=os.environ["TICKTICK_REDIRECT_URI"],  # TODO: set in env
            cache_path=f"{ROOT_DIR}/{cache_file}",
        )
        self.api = TickTick(username=os.environ["TICKTICK_EMAIL"],
                            password=os.environ["TICKTICK_PASSWORD"],
                            oauth=auth_client)
        self.projects = {project["id"]: project["name"] for project in self.api.state["projects"]}
        self.log.info(f"TickTickClient initialized with projects: {self.projects.values()}")

    def _get_project_name(self, project_id: str) -> str:
        """Return the project name corresponding to the given project id"""
        return self.projects[project_id] if project_id in self.projects else "Inbox"

    def _get_project_id(self, project_name: str) -> str:
        """Return the project id corresponding to the given project name"""
        for project_id, name in self.projects.items():
            if name == project_name:
                return project_id
        raise ValueError(f"Project {project_name} not found in TickTick")

    def _convert_date(self, date: str | datetime.date | datetime.datetime | None) -> str | None:
        """Convert a date to the TickTick format, if it isn't already"""
        if date is None or date == "" or pd.isna(date):
            return None
        elif isinstance(date, datetime.date) or isinstance(date, datetime.datetime):
            return date.strftime("%Y-%m-%dT%H:%M:%S+0000")
        elif isinstance(date, str) and len(date) == 10:
            return date + "T00:00:00+0000"
        elif isinstance(date, str) and len(date) == 24:
            return date
        else:
            raise TypeError(f"Date seems to be misconfigured, check type {type(date)} and len: 24")

    def create_task(
        self,
        title: str,
        project: str,
        content: str = "",
        start_date=None,
        due_date=None,
        all_day=True,
    ) -> str:
        """Create a task with the given title and project name, and optional content & due date

        Currently only supports one date in ticktick, to show in the "today" list. Therefore, tasks
        are created with the start date as the due date.

        Args:
            title: Title of the task
            project: Name of the project to create the task in
            content: Content of the task. Defaults to "".
            due_date: Due date of the task. Defaults to None.
                Will be converted to format: "2020-07-06T02:30:00+0100"
        """
        start_date = self._convert_date(start_date)
        due_date = self._convert_date(due_date)
        new_task = {
            "title": title,
            "content": content,
            "startDate": start_date,
            "dueDate": due_date,
            "isAllDay": all_day,
            "projectId": self._get_project_id(project),
        }
        try:
            new_task_result = self.api.task.create(task=new_task)
        except RetryError:
            raise RetryError("Apparent API throttling, could not create task")
        # self.log.debug(f"Created task: {new_task_result}")
        return new_task_result["id"]

    def create_tasks(self, titles: list[str], start_dates: list[str], project: str) -> list[str]:
        """Create tasks in TickTick from a given project. Return the ids of the created tasks.

        Expects a dataframe with the following columns:
            title: str, title of the task
            start_date: str|datetime|nan, start date of the task
            due_date: str|datetime|nan, due date of the task
        """
        assert len(titles) == len(start_dates), "titles and start_dates must be of equal length"
        self.log.info(f"Creating {len(titles)} tasks in project: {project}")
        created_ids = []
        for title, start_date in zip(titles, start_dates):
            created_id = self.create_task(
                title=title,
                start_date=start_date,
                # due_date=task["due_date"],
                project=project,
            )
            created_ids.append(created_id)
        return created_ids

    def _fix_timezones(self, tasks: DataFrame) -> DataFrame:
        """Checks if any dates are not 00:00 and add the timezone offset from the config.

        Converts to datetime, adds the offset, and converts back to string (can be improved)"""

        tasks["startDate"] = pd.to_datetime(tasks["startDate"], format=self.date_format)

        wrong_dates = tasks[(tasks["isAllDay"] == True) & tasks["startDate"].isna() == False]  # noqa
        wrong_dates = wrong_dates[
            (wrong_dates["startDate"].dt.hour != 0) | (wrong_dates["startDate"].dt.minute != 0)
        ]

        # add the timezone offset to the tasks that are in wrong_dates
        tasks.loc[wrong_dates.index, "startDate"] = wrong_dates["startDate"] + timedelta(
            hours=self.tz_offset[0], minutes=self.tz_offset[1]
        )
        tasks["startDate"] = tasks["startDate"].dt.strftime(self.date_format)
        tasks["startDate"] = tasks["startDate"].str[:-8] + tasks["startDate"].str[-5:]

        return tasks

    def get_all_tasks(self, project=None) -> DataFrame:
        """Return all tasks from TickTick as a DataFrame. Optionally, filter by project

        The dataframe is in the original 'TickTick' format."""
        tasks_df = DataFrame()
        for task in self.api.state["tasks"]:
            task["project"] = self._get_project_name(task["projectId"])
            tasks_df = tasks_df._append(task, ignore_index=True)
        if project is not None:
            tasks_df = tasks_df[tasks_df["project"] == project]
        self.log.info(f"Loaded {len(tasks_df)} tasks from TickTick (project: {project})")
        tasks_df = self._fix_timezones(tasks_df)
        return tasks_df

    def change_task_titles(self, task_ids: str | list[str], new_titles: str | list[str]) -> None:
        """Change the titles of a list of tasks"""
        if isinstance(task_ids, str) and isinstance(new_titles, str):
            task_ids = [task_ids]
            new_titles = [new_titles]
        assert len(task_ids) == len(new_titles), "task_ids and new_titles must be of equal length"
        for task_id, new_title in zip(task_ids, new_titles):
            task = self.api.get_by_id(task_id)
            task["title"] = new_title
            self.api.task.update(task)
            self.log.debug(f"Changed title of task {task_id} to {new_title}")

    def change_task_dates(self, task_ids: str | list[str], new_dates: str | list[str]) -> None:
        """Change the start dates of a list of tasks"""
        if isinstance(task_ids, str) and isinstance(new_dates, str):
            task_ids = [task_ids]
            new_dates = [new_dates]
        assert len(task_ids) == len(new_dates), "task_ids and new_dates must be of equal length"
        for task_id, new_date in zip(task_ids, new_dates):
            task = self.api.get_by_id(task_id)
            task["startDate"] = self._convert_date(new_date)
            self.api.task.update(task)
            self.log.debug(f"Changed start date of task {task_id} to {new_date}")

    def delete_all_tasks_in_project(self, project: str) -> None:
        """From a given project, delete all tasks"""
        project_tasks = self.get_all_tasks(project=project)
        self.log.info(f"Deleting all tasks in project: {project}")
        task_list = project_tasks.to_dict(orient="records")
        self.api.task.delete(task_list)

    def delete_tasks(self, task_ids: str | list[str]) -> None:
        # get the task objects via their ids
        if isinstance(task_ids, str):
            task_ids = [task_ids]
        self.log.info(f"Deleting {len(task_ids)} tasks")
        tasks = [self.api.get_by_id(task_id) for task_id in task_ids]
        self.api.task.delete(tasks)


if __name__ == "__main__":
    ticktick = TickTickClient()
    all_tasks = ticktick.get_all_tasks()
    notion_tasks = all_tasks[all_tasks["project"] == "Notion"]
    print(all_tasks)


from common_utils.web.cookies_handler import CookiesManager, LoginData, LoginSelectors


url = "https://web.budgetbakers.com/records"


class BudgetBakerHandler:
    cookies_manager = CookiesManager(
        login_data=LoginData(
            username_env="BUDGETBAKER_EMAIL",
            password_env="BUDGETBAKER_PASSWORD",
            sign_in_url="https://web.budgetbakers.com/login",
            selectors=LoginSelectors(
                username='input[name=email]',
                password='input[name=password]',
                login_button='button'
            )
        ),
        test_cookies_url="https://web.budgetbakers.com/records",
        test_cookies_response_fn=lambda data: "errorCode" not in data,
        min_num_cookies=5
    )

    def __init__(self):
        self.cookies = self.cookies_manager.get_cookies()
        self.headers = {"Cookie": self.cookies}

        raise NotImplementedError



import logging
import sys
import os
from logging.handlers import RotatingFileHandler


format_str = "[%(asctime)s]  %(levelname)s | %(name)s   -   %(message)s"
_long_date_format = "%Y-%m-%d %H:%M:%S"
_short_date_format = "%H:%M:%S"


if os.getenv("IS_CLOUD", False):
    import google.cloud.logging
    client = google.cloud.logging.Client()
    client.get_default_handler()
    client.setup_logging()
    format_str = "%(levelname)s |  %(name)s   -   %(message)s"


class FixedWidthFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None, name_length=16):
        super().__init__(fmt, datefmt)
        self.name_length = name_length

    def format(self, record, name_length=16):
        record.levelname = f"{record.levelname[:7]:<7}"
        record.name = f"{record.name[:name_length]:<{name_length}}"
        return super().format(record)


def create_logger(name: str, long_date_format=False, name_length=18) -> logging.Logger:
    """Creates a logger using a preset format str. Logs to "All"- and {name}.log if APPDATA_DIR"""
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    date_format = _long_date_format if long_date_format else _short_date_format
    log_formatter = FixedWidthFormatter(format_str, datefmt=date_format, name_length=name_length)
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setLevel(logging.DEBUG)
    stream_handler.setFormatter(log_formatter)
    logger.addHandler(stream_handler)

    if os.getenv("APPDATA", False) and os.getenv("APPDATA_DIR", False):
        log_folder = f"{os.getenv('APPDATA')}/{os.getenv('APPDATA_DIR')}/logs"
        os.makedirs(log_folder, exist_ok=True)
        log_file_all = f"{log_folder}/All.log"
        log_file = f"{log_folder}/{name}.log"
        file_handler_a = RotatingFileHandler(log_file_all, maxBytes=50 * 1024 * 1024, backupCount=5)
        file_handler_name = RotatingFileHandler(log_file, maxBytes=50 * 1024 * 1024, backupCount=5)
        file_handler_a.setFormatter(log_formatter)
        file_handler_a.setLevel(logging.DEBUG)
        file_handler_name.setFormatter(log_formatter)
        file_handler_name.setLevel(logging.DEBUG)
        logger.addHandler(file_handler_a)
        logger.addHandler(file_handler_name)

    return logger



import requests
import json


def get_request(url, headers=None, catch_exception=False) -> dict | None:
    """
    Send a POST request to the given URL with the given data and headers.
    """
    if headers is None:
        headers = {"Content-Type": "application/json"}
    try:
        response = requests.get(url, headers=headers)
        return response.json()
    except Exception as e:
        if catch_exception:
            return None
        else:
            raise Exception(f"Error sending POST request to {url}: {e}")



def post_request(url, data, headers=None, catch_exception=False) -> dict | None:
    """
    Send a POST request to the given URL with the given data and headers.
    """
    if headers is None:
        headers = {"Content-Type": "application/json"}
    try:
        response = requests.post(url, data=json.dumps(data), headers=headers)
        return response.json()
    except Exception as e:
        if catch_exception:
            return None
        else:
            raise Exception(f"Error sending POST request to {url}: {e}")



from dataclasses import dataclass
from typing import Literal
from time import sleep
from selenium.webdriver.common.by import By
from selenium import webdriver

# from undetected_chromedriver.v2 import Chrome as UndetectedChrome

from common_utils.logger import create_logger


@dataclass
class DriverAction:
    action: Literal["url", "sleep", "click", "send_keys", "get_text", "get_texts"]
    identifier: str | None = None
    input: str | int | None = None



class SeleniumHandler:
    log = create_logger("SeleniumHandler")

    def __init__(self, headless: bool = True, undetected: bool = False, verbose: bool = False, raise_exceptions: bool = True):
        self.headless = headless
        self.undetected = undetected
        self.verbose = verbose
        self.raise_exceptions = raise_exceptions

    def log_msg(self, message: str, level: Literal["debug", "info", "warning", "error"] = "info"):
        if self.verbose is False:
            return
        match level:
            case "debug":
                self.log.debug(message)
            case "info":
                self.log.info(message)
            case "warning":
                self.log.warning(message)
            case "error":
                self.log.error(message)
            case _:
                self.log.info(message)

    def get_driver(self):
        if self.undetected:
            raise NotImplementedError("Undetected Chrome is not supported in this version")
            # return UndetectedChrome()
        if self.headless:
            options = webdriver.ChromeOptions()
            options.add_argument("--headless")
            options.add_argument("--disable-gpu")
            options.add_argument("--no-sandbox")
            options.add_argument("--disable-dev-shm-usage")
            return webdriver.Chrome(options=options)
        return webdriver.Chrome()

    def run_actions(self, actions: list[DriverAction], driver: webdriver.Chrome | None = None):
        _driver = driver if driver is not None else self.get_driver()
        results = []

        for action in actions:
            match action.action:
                case "url":
                    _driver.get(action.identifier)
                    self.log_msg(f"Opened URL: {action.identifier}", level='debug')
                case "send_keys":
                    _driver.find_element(By.CSS_SELECTOR, action.identifier).send_keys(action.input)
                    self.log_msg(f"Sent keys: {action.input} to {action.identifier}", level='debug')
                case "click":
                    _driver.find_element(By.CSS_SELECTOR, action.identifier).click()
                    self.log_msg(f"Clicked on: {action.identifier}", level='debug')
                case "sleep":
                    self.log_msg(f"Sleeping for: {action.input} seconds", level='debug')
                    sleep(action.input)
                case "get_text":
                    result = _driver.find_element(By.CSS_SELECTOR, action.identifier).text
                    results.append(result)
                    self.log_msg(f"Storing text: {result} at idx {len(results)-1}", level='debug')
                case "get_texts":
                    result = [element.text for element in _driver.find_elements(By.CSS_SELECTOR, action.identifier)]
                    results.append(result)
                    self.log_msg(f"Storing texts: {result} at idx {len(results)-1}", level='debug')
                case _:
                    self.log_msg(f"Invalid action: {action.action}", level='error')
                    if self.raise_exceptions:
                        raise ValueError(f"Invalid action: {action.action}")

        if driver is None:
            _driver.quit()

        return results


from dataclasses import dataclass
import threading
from datetime import datetime
import random
import telebot
from telebot import TeleBot
from telebot.types import InlineKeyboardMarkup, InlineKeyboardButton


@dataclass
class QueryQuestion:
    question: str
    button_rows: list[list[tuple[str, str] | str]]

    def __repr__(self):
        return f'QueryQuestion({self.question})'


def poll_bot_threaded(bot: TeleBot):
    thread = threading.Thread(target=bot.polling)
    thread.start()


def query_multiple_questions(
    bot: TeleBot,
    chat_id: int,
    questions_data: list[QueryQuestion],
    callback_name: str | None = None,
    callback_func: callable = None,
    prefix_call_values: list[str] | None = None,
    separator: str = '~',
):
    if len(questions_data) < 1:
        return
    if callback_name is None:
        callback_name = str(datetime.now().timestamp()) + str(random.randint(0, 1000))
    prefix_call_values = prefix_call_values or []
    callback_data_prefix = callback_name + separator + separator.join(prefix_call_values)
    query_single_question(
        bot=bot,
        chat_id=chat_id,
        question_data=questions_data[0],
        callback_data_prefix=callback_data_prefix,
        separator=separator
    )
    _define_query_callback_handler(
        bot=bot,
        callback_name=callback_name,
        questions_data=questions_data,
        callback_func=callback_func,
        num_prefixes=len(prefix_call_values),
        separator=separator
    )


def query_single_question(
    bot: TeleBot,
    chat_id: int,
    question_data: QueryQuestion,
    callback_data_prefix: str,
    separator='~',
    message_id=None
):
    keyboard = InlineKeyboardMarkup()
    for button_row in question_data.button_rows:
        button_objects = []
        for button_data in button_row:
            text = button_data[0] if isinstance(button_data, tuple) else button_data
            callback = button_data[1] if isinstance(button_data, tuple) else button_data
            button_callback_data = callback_data_prefix + separator + callback
            button_object = InlineKeyboardButton(
                text.capitalize(), callback_data=button_callback_data
            )
            button_objects.append(button_object)
        keyboard.row(*button_objects)
    if message_id is None:
        bot.send_message(chat_id, question_data.question, reply_markup=keyboard)
    else:
        bot.edit_message_text(question_data.question, chat_id, message_id, reply_markup=keyboard)


def _define_query_callback_handler(
    bot: TeleBot,
    callback_name: str,
    questions_data: list[QueryQuestion],
    callback_func: callable,
    num_prefixes: int,
    separator='~'
):
    @bot.callback_query_handler(func=lambda call: call.data.startswith(callback_name))
    def query_callback_handler(call: telebot.types.CallbackQuery):
        callback_data = call.data.split(separator)
        query_answers = callback_data[num_prefixes + 1:]
        if len(query_answers) < len(questions_data):
            next_question_data = questions_data[len(query_answers)]
            query_single_question(
                bot=bot,
                chat_id=call.message.chat.id,
                question_data=next_question_data,
                callback_data_prefix=call.data,
                separator=separator,
                message_id=call.message.message_id
            )
        elif callback_func is not None:
            callback_func(callback_data[1:], message_id=call.message.id, chat_id=call.message.chat.id)



if __name__ == '__main__':
    def print_results(results):
        print(results)
    from dotenv import load_dotenv
    import os
    load_dotenv()
    chat_id_ = 784020584
    bot_ = telebot.TeleBot(os.getenv('TELEGRAM_BOT_TOKEN'))
    poll_bot_threaded(bot_)
    questions_data_ = [
        QueryQuestion(
            question='What is your favorite color?',
            button_rows=[
                ['red', 'green', 'blue'],
                ['yellow', 'purple', 'orange']
            ]
        ),
        QueryQuestion(
            question='What is your favorite animal?',
            button_rows=[
                ['dog', 'cat', 'elephant'],
                ['lion', 'tiger', 'bear']
            ]
        )
    ]
    query_multiple_questions(bot_, chat_id_, 'query', questions_data_, print_results, ['test1', 'tesgt23'])

